"""Main Entry point to masonry"""

import json
import os
from pathlib import Path

from .cli import parse_and_validate_args, parse_project_argument
from .prompt import prompt_add_template, prompt_init_project
from .template import add_template, initialise_project
from .utils import load_application_data, setup_logger

from.check import check_templates


def main(args=None):

    # Read in Arguments
    args = parse_and_validate_args(args)

    if args['-v']:
        logger = setup_logger(logfile=None, level='DEBUG')
    else:
        logger = setup_logger(logfile=None, level='INFO')

    logger.debug(f'Passed Arguments: {args}')

    project_templates = load_application_data()

    if args['init']:

        template = None

        # Get project
        if not args['PROJECT']:
            if project_templates:
                # Launch inquire
                project_dir = prompt_init_project(project_templates)
            else:
                raise ValueError("No known projects to choose from. "
                                 "Add one using the PROJECT argument.")
        else:
            project_dir, template = parse_project_argument(args['PROJECT'])

        # intialise a new project based on template
        output_dir = Path(args['--output']).resolve()

        initialise_project(
            project=project_dir,
            template=template,
            output_dir=output_dir)

    elif args['add']:

        project_dir = Path(args['--output']).resolve()

        # Load existing state information
        mason_vars = project_dir / '.mason'

        if not mason_vars.exists():
            raise IOError(f'A ".mason" file was not detected in the output directoty {project_dir}')

        with mason_vars.open('r') as f:
            project_state = json.load(f)

        project_templates_root = project_templates[project_state['project']]

        # Get project template options
        if not args['TEMPLATE']:
            # inquire which template to use
            args['TEMPLATE'] = prompt_add_template(project_templates_root, project_state)

        # Add the right template
        add_template(templates=args['TEMPLATE'],
                     project_dir=args['--output'])

    elif args['check']:

        template = None

        # Find project to test
        if not args['PROJECT']:
            if project_templates:
                # Launch inquire
                template_collection_path = prompt_init_project(project_templates)
            else:
                raise ValueError("No known projects to choose from. "
                                 "Add one using the PROJECT argument.")
        else:
            project_dir, template = parse_project_argument(args['PROJECT'])

        # Check all template layers
        # Add the right template
        check_templates(template_collection_path=template_collection_path,
                        template=template)


if __name__ == '__main__':
    main()




import json

import py.path

from masonry.resolution import create_dependency_graph, resolve
from masonry.template import add_template, initialise_project

from contextlib import redirect_stdout
from clint.textui import puts, indent, colored


def check_templates(template_collection_path, template=None, logfile='log'):

    default_template = get_default_template(template_collection_path)

    template_orders = get_template_orders(template_collection_path)

    tmpdir = py.path.local.mkdtemp().mkdir('check')

    # Rediect stdout to a file temporarily
    logfile = open(logfile, 'a')

    if default_template in template_orders:
        del template_orders[default_template]

    error_log = []

    for layer, deps in template_orders.items():

        if default_template in deps:
            deps.remove(default_template)

        output_dir = tmpdir.mkdir(layer)
        puts(colored.yellow(f'Testing Layer: {layer}'))
        puts(colored.yellow(f'Installing: {deps}'))

        project_dir = initialise_project(template_collection_path,
                                         output_dir=str(output_dir),
                                         interactive=False,
                                         stream=logfile.write)
        for dep in deps:
            try:
                puts(f'Adding layer: {dep}')
                add_template([dep], project_dir, interactive=False, stream=logfile.write)
            except Exception as e:
                msg = f'{e} encountered on adding "{dep}" to project for layer "{layer}"'
                error_log.append(msg)

    if error_log:
        puts(colored.red("Errors were encountered:"))
        with indent(4):
            for err in error_log:
                puts(colored.red(err))
    else:
        puts(colored.green("All template layers created without error"))


def get_template_orders(template_collection_path):
    """Return the orders in which to apply templates for all layers"""

    template_collection_path = py.path.local(template_collection_path)

    metadata = template_collection_path.join('metadata.json')

    layers = [x.basename for x in template_collection_path.listdir()
              if x.isdir() and x.basename != '__pycache__']

    g = create_dependency_graph(metadata, layers)

    orders = {}
    for layer in layers:
        orders[layer] = [x.name for x in resolve(g[layer])]

    return orders


def get_default_template(template_collection_path):

    template_collection_path = py.path.local(template_collection_path)

    # Find default template
    metadata = json.load(template_collection_path.join('metadata.json'))
    default_template = metadata['default']

    return default_template



""" Hook to add all items in the dev_environment file to the conda meat.ymal run requirements """

import pathlib
from ruamel.yaml import YAML

PY_VERSON = "python={{cookiecutter.python_version}}"

# Define Paths
environment_file = pathlib.Path('.').resolve() / 'dev_environment.yml'
conda_metadata_file = (
    pathlib.Path('.').resolve() / 'recipes' /
    '{{cookiecutter.package_name}}' / 'meta.yaml'
)

# init parser
yaml = YAML()
yaml.default_flow_style = False

# Read in files
environment_spec = yaml.load(environment_file.open())
conda_recipe = yaml.load(conda_metadata_file.open())

# Update
for package in environment_spec['dependencies']:
    if isinstance(package, str):
        if not package.startswith(PY_VERSON):
            conda_recipe['requirements']['run'].append(package)

# Write out
yaml.dump(conda_recipe, conda_metadata_file.open(mode='w'))





import pytest
from pathlib import Path

TEST_DIR = Path(__file__).parent


@pytest.fixture(autouse=True)
def no_prompts(monkeypatch):

    import inquirer

    def mock_inquirer_prompt(questions):
        answers = {q.name: q.default for q in questions}
        return answers

    # Mock the interactive input and use cookiecutter defaults instead
    monkeypatch.setattr(inquirer, 'prompt', mock_inquirer_prompt)




import json
import logging
import os
import shlex
import subprocess
from itertools import dropwhile
from pathlib import Path

from clint.textui import colored, indent, puts


def load_application_data(location=None):
    """Return the stored application data on where templates live"""

    if not location:
        location = os.getenv("HOME")

    app_data_dir = Path(os.path.join(location, ".masonry"))

    template_metadata_path = app_data_dir / "templates.json"

    if template_metadata_path.exists():
        obj = json.loads(template_metadata_path.read_text())
    else:
        obj = {}

    return obj


def save_application_data(obj, location=None):
    """Save application data on where templates live"""

    if not location:
        location = os.getenv("HOME")

    app_data_dir = Path(os.path.join(location, ".masonry"))
    app_data_dir.mkdir(exist_ok=True)

    template_metadata_path = app_data_dir / "templates.json"

    json.dump(obj, template_metadata_path.open('w'), indent=4)


def rindex(lst, item):
    """Return the index position of the last item in a list."""
    def index_ne(x):
        return lst[x] != item
    try:
        return next(dropwhile(index_ne, reversed(range(len(lst)))))
    except StopIteration:
        raise ValueError("rindex(lst, item): item not in list")


def setup_logger(name=__name__, logfile=None, level=logging.DEBUG):
    """
    A utility function that you can call to easily set up logging to the
    console and optionally to a file. No hassles.
    """
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(level)

    # Remove old handlers to allow updating settings
    for handler in list(logger.handlers):
        logger.removeHandler(handler)

    # create console handler
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)  # propagate all messages

    # add the formatter to the handler
    formatter = logging.Formatter('%(name)s - %(asctime)-15s - %(levelname)s: %(message)s')
    stream_handler.setFormatter(formatter)

    # setup logger and add the handlers
    logger.addHandler(stream_handler)

    if logfile:
        filehandler = logging.FileHandler(logfile)
        filehandler.setLevel(logging.NOTSET)
        filehandler.setFormatter(formatter)
        logger.addHandler(filehandler)

    logger.debug("logger set up. level=%s", level)
    return logger


def run_and_capture(command, give_feedback=True):

    p = subprocess.run(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if p.returncode != 0 and give_feedback:
        puts(colored.red('Error in executing "%s"' % command))
        puts(colored.red(p.stderr.decode().strip()))
    elif p.returncode == 0 and give_feedback:
        puts(colored.green('Executed "%s"' % command))

    return p



"""Handle dependeny resolutions"""
import json


class Node:
    """A simple class to represent a node in a graph"""

    def __init__(self, name):
        self.name = name
        self.edges = []

    def add_edge(self, node):
        self.edges.append(node)


def create_dependency_graph(filename, node_list):
    """ Return a dictionary of Node instances representing the Dependency Graph """

    meta_data = json.load(open(filename))
    default_template = meta_data["default"]

    # Initialise nodes from passed in node_list names
    nodes = {n: Node(n) for n in node_list}
    if default_template not in nodes:
        nodes[default_template] = Node(default_template)

    # Add in edges if dependencies are listed in the meta_data
    for k, v in nodes.items():
        deps = meta_data['dependencies'].get(k, None)
        if deps:
            for d in deps:
                v.add_edge(nodes[d])

    return nodes


def resolve(node, resolved=None, seen=None):
    """Graph dependency resolution algorithm

    Parameters
    ----------
    node: Node
        The node instance for which dependencies should be resolved

    Referenes
    ---------

        Source code obtained form the following paper. 

    .. https://www.electricmonk.nl/docs/dependency_resolving_algorithm/dependency_resolving_algorithm.html

    """
    if resolved is None:
        resolved = []
    if seen is None:
        seen = []

    seen.append(node)
    for edge in node.edges:
        if edge not in resolved:
            if edge in seen:
                raise Exception('Circular reference detected: %s -&gt; %s' % (node.name, edge.name))
            resolve(edge, resolved, seen)
    resolved.append(node)

    return resolved




from pathlib import Path
import pytest
import git
import json
import os

from conftest import TEST_DIR


@pytest.fixture(scope='module')
def init_base_project(tmpdir_factory):

    # Setup a basic project
    temp_output_path = Path(tmpdir_factory.mktemp('base_package').strpath)
    template_path = TEST_DIR / 'example_templates' / 'python_project'

    # Set arguments
    args = f"init -o {temp_output_path} {template_path}"

    from masonry import main
    # Run from entry point
    main.main(args=args)

    cookiecutter_vars_path = os.path.join(template_path, "package", "cookiecutter.json")
    with open(cookiecutter_vars_path, 'r') as f:
        cookiecutter_vars = json.load(f)

    package_name = cookiecutter_vars['package_name']
    project_dir = os.path.join(temp_output_path, package_name)

    return project_dir


def test_adding_single_project(init_base_project):

    current_project_path = Path(init_base_project)

    # Set arguments
    args = f"add -o {current_project_path.as_posix()} pytest"

    from masonry import main
    # Run from entry point
    main.main(args=args)

    # Check files were created
    files = [
        '.git/',
        '.mason',
        'MANIFEST.in',
        'README',
        'requirements.txt',
        'setup.py',
        'src/testpackage',
        'src/testpackage/__init__.py',
        'src/testpackage/main.py',
        'tests/test_foo.py'
    ]
    for f in files:
        p = current_project_path / f
        assert p.exists()

    # Check requirements were postfixed
    target = "requests\nlogzero\npytest\npytest-cov\ncoverage\n"
    req_file = current_project_path / 'requirements.txt'
    result = req_file.read_text()

    assert result == target

    # Check MANIFEST was prefixed
    target = "graft tests\ngraft src\n"
    manifest_file = current_project_path / 'MANIFEST.in'
    result = manifest_file.read_text()

    assert result == target

    # Check git repo was created and commits made
    repo_dir = current_project_path
    r = git.Repo(repo_dir.as_posix())
    log = r.git.log(pretty='oneline').split('\n')
    assert len(log) == 2
    assert "Add 'pytest' template layer via stone mason." in log[0]
    assert "Add 'package' template layer via stone mason." in log[1]


def test_adding_multiple_projects(init_base_project):

    current_project_path = Path(init_base_project)

    # Set arguments
    args = f"add -o {current_project_path.as_posix()} conda"

    from masonry import main
    # Run from entry point
    main.main(args=args)

    # Check files were created
    files = [
        '.git/',
        '.mason',
        'MANIFEST.in',
        'README',
        'requirements.txt',
        'setup.py',
        'src/testpackage',
        'src/testpackage/__init__.py',
        'src/testpackage/main.py',
        'tests/test_foo.py',
        'recipes/testpackage/build.bat',
        'recipes/testpackage/build.sh',
        'recipes/testpackage/run_test.py',
        'recipes/testpackage/meta.yaml'
    ]
    for f in files:
        p = current_project_path / f
        assert p.exists()

    # Check requirements were polulated
    target = "requests\nlogzero\npytest\npytest-cov\ncoverage\n"
    req_file = current_project_path / 'requirements.txt'
    result = req_file.read_text()

    assert result == target

    # Check MANIFEST was prefixed
    target = "graft tests\ngraft src\n"
    manifest_file = current_project_path / 'MANIFEST.in'
    result = manifest_file.read_text()

    assert result == target

    # Check git repo was created and commits made
    repo_dir = current_project_path
    r = git.Repo(repo_dir.as_posix())
    log = r.git.log(pretty='oneline').split('\n')
    assert len(log) == 3
    assert "Add 'conda' template layer via stone mason." in log[0]
    assert "Add 'pytest' template layer via stone mason." in log[1]
    assert "Add 'package' template layer via stone mason." in log[2]




import json
import shutil
import sys
import tempfile
from pathlib import Path

import git
from clint.textui import colored, prompt, puts, validators, indent

from .postprocess import combine_file_snippets
from .prompt import prompt_cookiecutter_variables
from .render import render_cookiecutter
from .resolution import create_dependency_graph, resolve
from .utils import load_application_data, save_application_data

STDOUT = sys.stdout.write


def initialise_project(project: Path, template=None, output_dir='.', interactive=True, stream=STDOUT):

    # Initialise path variables
    project_path = Path(project).resolve()
    meta_data_path = project_path / 'metadata.json'
    project_templates = load_application_data()

    # If project is a valid project file
    if not meta_data_path.exists():
        raise IOError("Not a valid project directory. Missing metadata.json file")

    if not template:
        # Load metadata for project and read in default template from there
        with meta_data_path.open() as meta_data_file:
            meta_data = json.load(meta_data_file)
        template = meta_data['default']

    # Store location of project template
    if project_path.name in project_templates:
        if project_templates[project_path.name] != project_path.as_posix():
            ans = prompt.query(
                "Project with the same name already exists in another location. Overwrite? [y/n]",
                validators=[validators.RegexValidator('[yn]')])
            if ans == 'n':
                sys.exit()

    project_templates[project_path.name] = project_path.as_posix()
    save_application_data(project_templates)

    # Work out all template names
    template_paths = {p.name: p for p in project_path.iterdir() if p.is_dir()}
    template_names = list(template_paths.keys())

    # Create graph of template dependencies
    g = create_dependency_graph(project_path / 'metadata.json',
                                node_list=template_names)

    # Resolve dependencies for specified template
    template_order = [n.name for n in resolve(g[template])]
    order_str = ' --> '.join(template_order)
    puts(colored.yellow(f'Creating new project from templates: {order_str}'), stream=stream)
    puts(stream=stream)

    # Initialise output structure to save state of project
    project_state = {}
    project_state['templates'] = []
    project_state['variables'] = {}

    # Cycle through templates and render them
    for name in template_order:
        template = template_paths[name].as_posix()
        if interactive:
            context_variables = prompt_cookiecutter_variables(template, project_state['variables'])
        else:
            context_variables_path = Path(template) / "cookiecutter.json"
            context_variables = json.load(context_variables_path.open())

        puts(f'Rendering "{name}" ...', stream=stream)

        output_project_dir, content = safe_render(
            template, output_dir, context=context_variables, stream=stream
        )

        # This combines any prefix/postfix files added by the template to the originally named file
        # e.g. the cntent of 'Makefile_postfix' with be added to the end of the file 'Makefile',
        # and the cntent of 'MANIFEST_pretfix.in' with be added to the start of the file 'MANIFEST.in'
        combine_file_snippets(output_project_dir, stream=stream)

        if not (Path(output_project_dir) / '.git').exists() and 'repo' not in vars():
            # Initialise git repo
            repo = git.Repo.init(output_project_dir)
        else:
            repo = git.Repo(output_project_dir)

        # Save state
        project_state['variables'].update(content)
        if name not in project_state['templates']:
            project_state['templates'].append(name)
        project_state['project'] = project_path.name

        # Save state of project variables
        mason_vars = Path(output_project_dir) / '.mason'
        with mason_vars.open('w') as f:
            json.dump(project_state, f, indent=4)

        # Commit template layer to git repo
        all_files = [p.as_posix() for p in Path(output_project_dir).iterdir() if p.is_file]
        repo.index.add(all_files)
        repo.index.commit(f"Add '{name}' template layer via stone mason.")

    return output_project_dir


def add_template(templates: list, project_dir: Path, interactive=True, stream=STDOUT):
    """ Add a template to an existing project """

    project_dir = Path(project_dir).resolve()
    repo = git.Repo(project_dir.as_posix())

    # Load existing state information
    mason_vars = project_dir / '.mason'
    with mason_vars.open('r') as f:
        project_state = json.load(f)

    # Load project locations and get project root path
    project_template_data = load_application_data()
    project_root = Path(project_template_data[project_state['project']])

    # Find remaining templates that can be applied
    previous_templates = project_state['templates']
    paths = project_root.iterdir()
    template_paths = {p.name: p for p in paths if p.is_dir()}

    templates_names = list(template_paths.keys())
    templates_names.sort()

    for t in templates:
        assert t in templates_names, f'{t} not in templates for this type of project'

    # Create graph of template dependencies
    g = create_dependency_graph(project_root / 'metadata.json',
                                node_list=templates_names)

    # Resolve dependencies for specified template
    template_orders = []

    for t in templates:
        order = [n.name for n in resolve(g[t]) if n.name not in previous_templates]
        if order:
            template_orders.append(order)
            previous_templates.extend(order)

    order_strs = [' --> '.join(order) for order in template_orders]

    puts(colored.yellow(f'Adding the following templates to project:'), stream=stream)
    with indent(4):
        for order in order_strs:
            puts(colored.yellow(order), stream=stream)
    puts(stream=stream)

    # Cycle through templates and render them
    for order_set in template_orders:

        for name in order_set:
            template = template_paths[name].as_posix()

            if interactive:
                context_variables = prompt_cookiecutter_variables(
                    template, project_state['variables'])
            else:
                context_variables_path = Path(template) / "cookiecutter.json"
                context_variables = json.load(context_variables_path.open())

            # Only update dict with new variables.
            for k, v in context_variables.items():
                if k not in project_state['variables']:
                    project_state['variables'][k] = v

            puts(f'Rendering "{name}" ...', stream=stream)

            output_project_dir, content = safe_render(
                template,
                target_dir=project_dir.parent,
                context=project_state['variables'],
                stream=stream
            )

            # This combines any prefix/postfix files added by the template to the originally named file
            # e.g. the cntent of 'Makefile_postfix' with be added to the end of the file 'Makefile',
            # and the cntent of 'MANIFEST_pretfix.in' with be added to the start of the file 'MANIFEST.in'
            combine_file_snippets(output_project_dir, stream=stream)

            # Save state
            project_state['variables'].update(content)
            if name not in project_state['templates']:
                project_state['templates'].append(name)

            # Save state of project variables
            with mason_vars.open('w') as f:
                json.dump(project_state, f, indent=4)

            output_project_dir = Path(output_project_dir)

            # Commit template layer to git repo
            repo.git.add(output_project_dir.as_posix(), force=False)
            repo.index.commit(f"Add '{name}' template layer via stone mason.")


def safe_render(template, target_dir, context, stream=STDOUT):
    """Safely Render a new template by first making a backup to roll back to if needed."""

    # Copy current directory to temp backup location
    backup_dir = Path(tempfile.mkdtemp()) / 'backup'
    backup = shutil.copytree(target_dir, backup_dir)

    try:
        output_dir, content = render_cookiecutter(
            template, no_input=True,
            extra_context=context,
            output_dir=target_dir,
            overwrite_if_exists=True,
        )
    except Exception as e:
        # Rollback
        puts(colored.red("An error occured during templating, Rolling back to last stable state."), stream=stream)
        shutil.rmtree(target_dir)
        shutil.copytree(backup, target_dir)
        with indent(4):
            puts(f'Restored {backup_dir} to {target_dir}', stream=stream)
            puts(colored.red(f"Traceback: \n{e}"), stream=stream)
        raise e
        sys.exit()

    # shutil.rmtree(backup_dir.as_posix())

    return output_dir, content



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='{{cookiecutter.package_name}}',
    description='{{cookiecutter.short_description}}',
    author='{{cookiecutter.author}}',
    author_email='{{cookiecutter.email}}',
    # Required for packaging source distribution from 'src' subfolder
    packages=find_packages(where='src'),
    package_dir={'': 'src'},

    # This next section registers 'mycommand' as a command line utility that will execute
    # the `cli_entry_point()` function in the `command_line` module of `mypackage`
    #entry_points={
    #    'console_scripts': [
    #        'mycommand=mypackage.main:cli_entry_point'
    #    ]
    #}
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Im a doc sting!!!
"""

import sys
import os

from fileinput import FileInput

# Here are some data science libraries
import pandas as pd
import numpy as np




from pkg_resources import get_distribution, DistributionNotFound
import pathlib

# Code to lookup version number as obtained from setuptools_scm
# https://pypi.python.org/pypi/setuptools_scm
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = None

# Store location of where the package is installed
PKG_DIR = pathlib.Path(__file__).parent



import json

import py.path
import pytest

from masonry.resolution import create_dependency_graph, resolve
from masonry.template import initialise_project, add_template


TEMPLATE_COLLECTION_PATH = (
    py.path.local(__file__)
    .dirpath()
    .join('example_templates/python_project')
)


@pytest.fixture
def template_orders():

    metadata = TEMPLATE_COLLECTION_PATH.join('metadata.json')
    layers = [x.basename for x in TEMPLATE_COLLECTION_PATH.listdir()
              if x.isdir() and x.basename != '__pycache__']

    g = create_dependency_graph(metadata, layers)

    orders = {}
    for layer in layers:
        orders[layer] = [x.name for x in resolve(g[layer])]

    return orders


@pytest.fixture
def default_template():

    # Find default template
    metadata = json.load(TEMPLATE_COLLECTION_PATH.join('metadata.json'))
    default_template = metadata['default']

    return default_template


def test_all_layers(template_orders, default_template, tmpdir):

    # Given the order of all dependencies for the template layers
    if default_template in template_orders:
        del template_orders[default_template]

    # When adding each in turn
    for layer, deps in template_orders.items():

        if default_template in deps:
            deps.remove(default_template)

        output_dir = tmpdir.mkdir(layer)
        print('Testing Layer:', layer)
        print('Initialising with default template')
        project_dir = initialise_project(TEMPLATE_COLLECTION_PATH, output_dir=str(output_dir))

        for dep in deps:
            print('Adding layer:', dep)
            add_template([dep], project_dir)




from pathlib import Path
import pytest
import git
import json

from conftest import TEST_DIR


def test_init_with_project(tmpdir):

    output_path = Path(tmpdir.strpath)

    # Set arguments
    args = f"init -o {output_path} {TEST_DIR}/example_templates/python_project"

    from masonry import main
    # Run from entry point
    main.main(args=args)

    # Check files were created
    package_name = 'testpackage'
    files = [
        '.git/',
        '.mason',
        'MANIFEST.in',
        'README',
        'requirements.txt',
        'setup.py',
        'src/testpackage',
        'src/testpackage/__init__.py',
        'src/testpackage/main.py'
    ]
    for f in files:
        p = output_path / package_name / f
        assert p.exists()

    # Check requirements were polulated
    target = "requests\nlogzero\n"
    req_file = output_path / package_name / 'requirements.txt'
    result = req_file.read_text()
    assert result == target

    # Check git repo was created and commits made
    repo_dir = output_path / package_name
    r = git.Repo(repo_dir.as_posix())
    log = r.git.log(pretty='oneline').split('\n')
    assert len(log) == 1
    assert "Add 'package' template layer via stone mason." in log[0]


def test_init_with_project_and_template(tmpdir, no_prompts):

    output_path = Path(tmpdir.strpath)

    # Set arguments
    args = f"init -o {output_path} {TEST_DIR}/example_templates/python_project/pytest"

    from masonry import main
    # Run from entry point
    main.main(args=args)

    # Check files were created
    package_name = 'testpackage'
    files = [
        '.git/',
        '.mason',
        'MANIFEST.in',
        'README',
        'requirements.txt',
        'setup.py',
        'src/testpackage',
        'src/testpackage/__init__.py',
        'src/testpackage/main.py',
        'tests/test_foo.py'
    ]
    for f in files:
        p = output_path / package_name / f
        assert p.exists()

    # Check requirements were polulated
    target = "requests\nlogzero\npytest\npytest-cov\ncoverage\n"
    req_file = output_path / package_name / 'requirements.txt'
    result = req_file.read_text()

    assert result == target

    # Check MANIFEST was prefixed
    target = "graft tests\ngraft src\n"
    manifest_file = output_path / package_name / 'MANIFEST.in'
    result = manifest_file.read_text()

    assert result == target

    # Check git repo was created and commits made
    repo_dir = output_path / package_name
    r = git.Repo(repo_dir.as_posix())
    log = r.git.log(pretty='oneline').split('\n')
    assert len(log) == 2
    assert "Add 'pytest' template layer via stone mason." in log[0]
    assert "Add 'package' template layer via stone mason." in log[1]



import inquirer

from pathlib import Path
import json

IGNORED_DIRS = [
    '__pycache__',
    '.git'
]


def prompt_init_project(project_templates):
    """Return the choice of project to intialise"""

    projects = list(project_templates.keys())
    projects.sort()

    questions = [
        inquirer.List('project',
                      message="What project do you want to initialise?",
                      choices=projects)
    ]

    answers = inquirer.prompt(questions)

    project_path = project_templates[answers['project']]

    return project_path


def prompt_add_template(project_template_root, project_state):
    """Return the choice of templates to add"""

    project_template_root = Path(project_template_root).resolve()

    # Find remaining templates that can be applied
    previous_templates = project_state['templates']

    paths = project_template_root.iterdir()

    remaining_template_paths = {
        p.name: p for p in paths
        if p.is_dir() and p.name not in previous_templates and p.name not in IGNORED_DIRS
    }

    print(remaining_template_paths.keys())

    remaining_templates_names = list(remaining_template_paths.keys())
    remaining_templates_names.sort()

    questions = [
        inquirer.Checkbox('templates',
                          message="What templates do you wish to add?",
                          choices=remaining_templates_names)
    ]

    answers = inquirer.prompt(questions)

    templates = answers['templates']

    return templates


def prompt_cookiecutter_variables(template, context_variables):

    template_path = Path(template)
    cookiecutter_vars_path = template_path / "cookiecutter.json"

    with cookiecutter_vars_path.open() as f:
        cookiecutter_vars = json.load(f)

    current_variables = list(context_variables.keys())

    # Construct list of questions for variables without a value
    questions = []
    for key, value in cookiecutter_vars.items():

        if key not in current_variables:

            var_name = key.replace("_", " ").title()

            if isinstance(value, (str, int, float, bool)):
                questions.append(
                    inquirer.Text(key,
                                  message=f"Please specify '{var_name}'",
                                  default=value)
                )
            elif isinstance(value, list):
                questions.append(
                    inquirer.List(key,
                                  message=f"Please specify '{var_name}'",
                                  default=value[0],
                                  choices=value)
                )

    answers = inquirer.prompt(questions)

    return answers



"""Masonry

A tool for working with composable project templates. 

Commands:
    init  - initialise a new project 
    add   - add a template to an existing project
    check - check a particular project or template layer renders correctly

Usage:
    mason init [-v] [-o DIR] [PROJECT]
    mason add [-v] [-o DIR] [TEMPLATE ...]
    mason check [-v] [PROJECT]
    mason (-h | --help)
    mason --version

Arguments:
  PROJECT      URL/Filepath to directory for project type. Should contain
               subdiectories for all templates and a metadata.json file.
               Can also specify a template subdirectory, and if so, this
               is used as the template instead of the default for the project.

  TEMPLATE     Name of one or more templates to add to an existing project.

Options:
  -o DIR --output=DIR  Project directory to create/add to [default: .]
  -h --help            Show this screen.
  --version            Show version.
  -v                   Increase verbosity of output
"""
import os
from sys import exit
from pathlib import Path

from docopt import docopt

from . import __version__

try:
    from schema import Schema, And, Or, Use, SchemaError
except ImportError:
    exit('This example requires that `schema` data-validation library'
         ' is installed: \n    pip install schema\n'
         'https://github.com/halst/schema')


def parse_args(argv=None):
    args = docopt(__doc__, version=__version__, argv=argv)
    return args


def validate_args(args):
    def template_path_exists(template):
        if template:
            template_path = os.path.join(args['PROJECT'], template)
            return os.path.exists(template_path)
        else:
            return True

    def mason_file_exists(output):
        mason_path = os.path.join(output, '.mason')
        return os.path.exists(mason_path)

    if args['init']:
        schema = Schema({
            # 'TEMPLATE': And(template_path_exists, error='TEMPLATE should be subdirectory of PROJECT'),
            # 'PROJECT': And(os.path.exists, error='PROJECT should exist'),
            str: object})

    elif args['add']:
        schema = Schema({
            # 'TEMPLATE': And(str, error='TEMPLATE should be specified'),
            # '--output': And(mason_file_exists, error='The .mason was not detected for project.'
            #                                          ' Specofy location with --output option.'),
            str: object})

    elif args['check']:
        schema = Schema({
            # 'TEMPLATE': And(str, error='TEMPLATE should be specified'),
            # '--output': And(mason_file_exists, error='The .mason was not detected for project.'
            #                                          ' Specofy location with --output option.'),
            str: object})
    try:
        args = schema.validate(args)
    except SchemaError as e:
        exit(e)


def parse_and_validate_args(argv):
    parsed_args = parse_args(argv)
    validate_args(parsed_args)
    return parsed_args


def parse_project_argument(arg):

    template = None

    # See if template directroy or project directory was specified
    project_argument = Path(arg).resolve()
    project_metadata = project_argument / 'metadata.json'

    if project_metadata.exists():
        project_dir = arg

    elif (project_argument.parent / 'metadata.json').exists():
        # Check parent directory and use sub directory as template
        project_dir = project_argument.parent.as_posix()
        template = project_argument.name

    return project_dir, template



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Im a doc sting!!!
"""

import sys
import os

from fileinput import FileInput

# Here are some data science libraries
import pandas as pd
import numpy as np

MY_CONSTANT = 12124

if __name__ == '__main__':
    print(function1(5, 6))

    print(function2(5, 6))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Im a doc sting!!!
"""

import sys
import os

from fileinput import FileInput

# Here are some data science libraries
import pandas as pd
import numpy as np

MY_CONSTANT = 12124


# This function is important
def function1(arg1, arg2):
    """This is my doc string of things"""
    ans = arg1 + arg2
    return ans


def function2(arg1, arg2):
    """This is my doc string of things"""
    ans = f'  {arg1} + {arg2}  '
    return ans


def function3(arg1, arg2):
    """This is my doc string of things

    Here is an
    extended decription.
    """
    ans = arg1 + arg2
    return ans


if __name__ == '__main__':
    print(function1(5, 6))

    print(function2(5, 6))



""" Tests for code and text insertion postprocessing
"""

import pytest
from pathlib import Path
import shutil
import itertools


from conftest import TEST_DIR

from masonry.postprocess import insert_code, combine_file_snippets


ns = range(1, 6)
kinds = ('pre', 'post')


@pytest.mark.parametrize('n,kind', itertools.product(ns, kinds))
def test_code_insertion(n, kind):

    # GIVEN original python code file and some new code to insert into it
    original_path = TEST_DIR / Path(f'data/postprocessing_examples/input/test_code{n}.py')
    new_code_path = TEST_DIR / Path(f'data/postprocessing_examples/input/new_code.py')
    target_path = TEST_DIR / Path(f'data/postprocessing_examples/output/result{n}_{kind}.py')

    # WHEN code is inserted
    result = insert_code(new_code_path, original_path, kind=f'{kind}fix')

    # THEN result should have the code in the right place, with pre and post being before or after
    # the main body of the code, which is the part after inports and constants, but before the
    # if __name__ == __main__ clause, if present
    target = target_path.read_text()
    assert target == result


def test_combine_file_snippets(tmpdir):

    # GIVEN a directory of files, with *_post and *_prefixes that match other named files.
    input_files = [
        TEST_DIR / Path('data/postprocessing_examples/input/config.ini'),
        TEST_DIR / Path('data/postprocessing_examples/input/config_postfix.ini'),
        TEST_DIR / Path('data/postprocessing_examples/input/main.py'),
        TEST_DIR / Path('data/postprocessing_examples/input/main_prefix.py')
    ]
    # Use a temporary directory for the processing
    dest_path = Path(tmpdir.strpath) / 'postprocessing'
    dest_path.mkdir()
    for fp in input_files:
        shutil.copy(str(fp), str(dest_path))

    # WHEN combining files using the postprocessing rules
    combine_file_snippets(tmpdir.strpath)

    # THEN main_prefix.py should be prefixed to main.py
    target_path = TEST_DIR / Path('data/postprocessing_examples/output/result1_pre.py')
    target = target_path.read_text()
    result_path = dest_path / 'main.py'
    result = result_path.read_text()
    assert target == result

    # THEN config_prefix.ini should be prefixed to config.ini
    target_path = TEST_DIR / Path('data/postprocessing_examples/output/result_config.ini')
    target = target_path.read_text()
    result_path = dest_path / 'config.ini'
    result = result_path.read_text()
    assert target == result

    # THEN *_prefix and *_postfix files hould be removed as part of the processing
    main_file_present = (result_path / 'main_prefix.py').exists()
    assert not main_file_present
    config_file_present = (result_path / 'config_postfix.py').exists()
    assert not config_file_present

    # CLEANUP
    shutil.rmtree(str(dest_path))




import {{cookiecutter.package_name}}
import pytest


def test_foo_does_x():

    assert False

def test_foo_does_y():

    assert False


# -------- #
# Examples #
# -------- #

# Tesing Exceptions 
# ---------------------------------------------------

# import pytest
# def f():
#     raise SystemExit(1)

# def test_mytest():
#     with pytest.raises(SystemExit):
#         f()

# Fixtures
# -----------------------------------------------------------
# Run `pytest -q --fixtures` to see a list of all builtin fixtures

# import pytest

# @pytest.fixture
# def smtp():
#     import smtplib
#     return smtplib.SMTP("smtp.gmail.com")

# def test_ehlo(smtp):
#     response, msg = smtp.ehlo()
#     assert response == 250
#     assert 0 # for demo purposes


# Monkey Patching
# ------------------------------------------------------------------

# import os.path
# def getssh(): # pseudo application code
#     return os.path.join(os.path.expanduser("~admin"), '.ssh')

# def test_mytest(monkeypatch):
#     def mockreturn(path):
#         return '/abc'
#     monkeypatch.setattr(os.path, 'expanduser', mockreturn)
#     x = getssh()
#     assert x == '/abc/.ssh'

# Parameterising tests
# -----------------------------------------------------------------------

# import pytest
# @pytest.mark.parametrize("test_input,expected", [
#     ("3+5", 8),
#     ("2+4", 6),
#     ("6*9", 42),
# ])
# def test_eval(test_input, expected):
#     assert eval(test_input) == expected


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'masonry'
DESCRIPTION = 'A command line tool for composable project templating.'
URL = 'https://github.com/MrKriss/masonry'
EMAIL = 'chris.j.musselle@gmail.com'
AUTHOR = 'Chris Musselle'

REQUIRED = [
    'cookiecutter>=1.6',
    'docopt',
    'schema',
    'inquirer',
    'ruamel.yaml>=0.15',
    'gitpython',
    'clint'
]

repo_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(repo_root, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(repo_root, 'package', 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


setup(
    # Metadata
    name=NAME,
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    # Use version control to define version tag
    use_scm_version={"root": "..", "relative_to": __file__},
    # Package Locations
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    # Dependencies
    install_requires=REQUIRED,
    include_package_data=True,

    # Entry Point for the command line application
    entry_points={
        'console_scripts': ['mason=masonry.main:main']
    },
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
    ],
    # Command for setup.py publishing support.
    cmdclass={
        'upload': UploadCommand,
    },
)




import ast
import fnmatch
import os

from .utils import rindex
from clint.textui import puts, indent, STDOUT


def combine_file_snippets(project_dir, stream=STDOUT):
    """ Recusively search for files ending in posfix/prefix and join them to their originals """

    for dirpath, dirnames, filenames in os.walk(project_dir):

        if '.git' in dirnames:
            dirnames.remove('.git')

        postfix_files = fnmatch.filter(filenames, '*_postfix*')
        prefix_files = fnmatch.filter(filenames, '*_prefix*')

        if postfix_files or prefix_files:
            puts(f"Applying postprocessing in {dirpath} ...", stream=stream)

        with indent(4):
            for postfile in postfix_files:
                original = os.path.join(dirpath, postfile.replace('_postfix', ''))

                if os.path.exists(original):
                    postfile = os.path.join(dirpath, postfile)

                    if postfile.endswith('.py') and original.endswith('.py'):
                        all_text = insert_code(postfile, original, kind='postfix')
                        with open(original, 'w') as f:
                            f.write(all_text)
                        os.remove(postfile)
                        puts(
                            f'Postfixing python code to {os.path.basename(original)}',
                            stream=stream)
                    else:
                        postfix_text(postfile, original)
                        puts(f'Postfixing text to {os.path.basename(original)}', stream=stream)
                else:
                    raise ValueError(f'Original file not found for specified postfix: {postfile}')

        with indent(4):
            for prefile in prefix_files:
                original = os.path.join(dirpath, prefile.replace('_prefix', ''))

                if os.path.exists(original):
                    prefile = os.path.join(dirpath, prefile)

                    if prefile.endswith('.py') and original.endswith('.py'):
                        all_text = insert_code(prefile, original, kind='prefix')
                        with open(original, 'w') as f:
                            f.write(all_text)
                        os.remove(prefile)
                        puts(
                            f'Prefixing python code to {os.path.basename(original)}', stream=stream)
                    else:
                        prefix_text(prefile, original)
                        puts(f'Prefixing text to {os.path.basename(original)}', stream=stream)
                else:
                    raise ValueError(f'Original file not found for specified prefix: {prefile}')


def postfix_text(src, dest):
    """ Add the content of src to the end of dest file and remove src file """

    with open(dest, 'a') as fout:
        with open(src) as fin:
            fout.write(fin.read())
    os.remove(fin.name)


def prefix_text(src, dest):
    """ Add the content of src file path to the begining of dest file path and remove src """

    with open(src, 'a') as fout:
        with open(dest) as fin:
            fout.write(fin.read())
    os.remove(fin.name)
    os.rename(fout.name, fin.name)


def insert_code(src, dest, kind):
    """Insert code in source into destination file."""

    source_text = open(src).read().strip()
    destination_text = open(dest).read()
    destination_lines = destination_text.split('\n')

    destination_tree = ast.parse(destination_text)

    if not destination_tree.body:
        idx = 0
    elif kind == "prefix":
        idx = find_prefix_insertion_idx(destination_tree)
    elif kind == "postfix":
        idx = find_postfix_insertion_idx(destination_tree)

    if idx >= len(destination_tree.body):

        # Strip blank line before insertion
        if destination_lines[-1].strip() == '':
            del destination_lines[-1]

        # Append to file
        destination_lines.append('\n\n' + source_text + '\n')

    else:

        # Start with index at first line above object definition
        line_no = destination_tree.body[idx].lineno - 1  # line numbers count from 1
        line_no = get_previous_blank_line_no(destination_lines, line_no)

        # Strip blank lines before insertion
        if destination_lines[line_no - 1].strip() == '':
            del destination_lines[line_no - 1]
            line_no -= 1

        # perform the insertion
        destination_lines.insert(line_no, '\n\n' + source_text + '\n')

    all_text = '\n'.join(destination_lines)

    return all_text


def find_prefix_insertion_idx(module_ast):
    """Return prefix insertion point. 

    This is the point in the code after any imports and constants, but before any functions or 
    classes are defined. 
    """

    # Inspect ast types present
    node_types = [type(x) for x in module_ast.body]
    func_class_pos = [x == ast.FunctionDef or x == ast.ClassDef for x in node_types]
    import_pos = [x == ast.Import or x == ast.ImportFrom for x in node_types]

    docstring_present = (isinstance(node_types[0], ast.Expr)
                         and isinstance(node_types[0].value, ast.Str))
    constants_pos = [isinstance(elem, ast.Assign) and elem.targets[0].id.isupper()
                     for elem in module_ast.body]

    last_node = module_ast.body[-1]
    ifname_present = isinstance(last_node, ast.If) and last_node.test.left.id == '__name__'

    if any(func_class_pos):
        # Insertion point is just before the first occurance of function or class
        insert_idx = func_class_pos.index(True)

    elif any(constants_pos):
        # Insertion point is after the last constant
        insert_idx = rindex(constants_pos, True)
        insert_idx += 1

    elif any(import_pos):
        # Insertion point is after all imports
        insert_idx = rindex(import_pos, True)
        insert_idx += 1

    elif docstring_present:
        # Insertion point is after docstring
        insert_idx = 1

    elif ifname_present:
        # Insertion point is before ifname
        insert_idx = -1

    return insert_idx


def find_postfix_insertion_idx(module_ast):
    """Return postfix insertion point. 

    This is the point in the code after any functions or classes are defined, but before the ifname
    statement if present. 
    """

    # Inspect ast types present
    node_types = [type(x) for x in module_ast.body]
    func_class_pos = [x == ast.FunctionDef or x == ast.ClassDef for x in node_types]

    last_node = module_ast.body[-1]
    ifname_present = isinstance(last_node, ast.If) and last_node.test.left.id == '__name__'

    if any(func_class_pos):
        # Insertion point is just after the last occurance of function or class
        insert_idx = rindex(func_class_pos, True)
        insert_idx += 1

    elif ifname_present:
        # Insertion point is before ifname
        insert_idx = -1

    else:
        # Insert at end of code
        insert_idx = len(module_ast.body)

    return insert_idx


def get_previous_blank_line_no(lines, idx):
    """Find a blank line before the object definition"""

    while True:
        if not lines[idx].strip():
            # found a blank line
            break
        else:
            idx -= 1

    return idx




from pathlib import Path
import pytest
import git
import json
import os

from conftest import TEST_DIR

from masonry import main

from cookiecutter.exceptions import FailedHookException, UndefinedVariableInTemplate


@pytest.fixture(scope='module')
def init_simple_project(tmpdir_factory):

    # Setup a basic project
    temp_output_path = Path(tmpdir_factory.mktemp('simple_project').strpath)
    template_path = TEST_DIR / 'example_templates' / 'breaking_project'

    # Set arguments
    args = f"init -o {temp_output_path} {template_path}"

    from masonry import main
    # Run from entry point
    main.main(args=args)

    cookiecutter_vars_path = os.path.join(template_path, "first_layer", "cookiecutter.json")
    with open(cookiecutter_vars_path, 'r') as f:
        cookiecutter_vars = json.load(f)

    project_name = cookiecutter_vars['project_name']
    project_dir = temp_output_path / project_name

    return project_dir


def test_rollback_when_error_in_pre_hook(init_simple_project):

    # GIVEN an initialised project
    project_dir = init_simple_project

    # WHEN a template is added that causes an error
    args = f"add -o {project_dir} breaking_pre_hook"
    with pytest.raises(FailedHookException):
        main.main(args=args)

    # THEN only the original files should be present
    target = set([
        project_dir / 'file_from_layer_1.txt',
        project_dir / '.mason',
        project_dir / '.git',
    ])
    result = set(project_dir.iterdir())
    assert result == target

    # THEN original file should be unchanged
    target = '123456'
    result_file = project_dir / 'file_from_layer_1.txt'
    result = result_file.read_text()
    assert result == target


def test_rollback_when_error_in_post_hook(init_simple_project):

    # GIVEN an initialised project
    project_dir = init_simple_project

    # WHEN a template is added that causes an error
    args = f"add -o {project_dir} breaking_post_hook"
    with pytest.raises(FailedHookException):
        main.main(args=args)

    # THEN only the original files should be present
    target = set([
        project_dir / 'file_from_layer_1.txt',
        project_dir / '.mason',
        project_dir / '.git',
    ])
    result = set(project_dir.iterdir())
    assert result == target

    # THEN original file should be unchanged
    target = '123456'
    result_file = project_dir / 'file_from_layer_1.txt'
    result = result_file.read_text()
    assert result == target


def test_rollback_when_error_in_variable_name(init_simple_project):

    # GIVEN an initialised project
    project_dir = init_simple_project

    # WHEN a template is added that causes an error
    args = f"add -o {project_dir} breaking_variable_name"
    with pytest.raises(UndefinedVariableInTemplate):
        main.main(args=args)

    # THEN only the original files should be present
    target = set([
        project_dir / 'file_from_layer_1.txt',
        project_dir / '.mason',
        project_dir / '.git',
    ])
    result = set(project_dir.iterdir())
    assert result == target

    # THEN original file should be unchanged
    target = '123456'
    result_file = project_dir / 'file_from_layer_1.txt'
    result = result_file.read_text()
    assert result == target


def test_rollback_when_init_project(tmpdir_factory):

    # GIVEN a temp directory and template to initialise
    temp_output_path = Path(tmpdir_factory.mktemp('empty_project').strpath)
    template_path = TEST_DIR / 'example_templates' / 'breaking_project'

    # WHEN a new project is initialised that causes an error
    args = f"init -o {temp_output_path} {template_path}/breaking_variable_name"
    with pytest.raises(UndefinedVariableInTemplate):
        main.main(args=args)

    # THEN the directory should be empty
    target = set([])
    result = set(temp_output_path.iterdir())
    assert result == target



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

MY_CONSTANT = 12124


def my_new_first_function(arg1, arg2):
    """This is my doc string of things"""
    ans = arg1 * arg2
    return ans


a = 10
b = 20

if __name__ == '__main__':
    print(function1(5, 6))

    print(function2(5, 6))
